
typedef int timerid_t;
typedef void fileproc_t(int fd, int mask, void*);
typedef void freeproc_t(void*);
typedef void timerproc_t(void*, timerid_t);
typedef void threadproc_t(void*);
typedef void sigproc_t(sigproc_t_arg);


/*
 *	Thread
 */
class sthread_t : private Thread  {
public:
    enum { rd = 1, wr = 2, ex = 4}; // read, write or exception flags
    enum { usr_size = 256 };
    static freeproc_t* const _static;
    static freeproc_t* const _delete;
    static freeproc_t* const _free;

    char	usr[usr_size]; // user can use this

    sthread_t(threadproc_t* func, void* arg, freeproc_t* f);
    ~sthread_t();

    /*
     *	thread controls
     */
    void thread_exit(int stat = 0);
    void yield();
    void block();
    void unblock();

    /*
     * 	status info
     */
    static num_threads();
    static num_ready();
    static sthread_t* me();

    /*
     *	debugging utilities
     */
    static void print_stat();
    static void dump();
    
    /*
     *	create shared memory segment
     */
    static void* create_buf_sgmt(int nbytes);
    
    /*
     *	overload unix system calls
     */
    _open(const char* path, int flags, int mode);
    _close(int fd);
    off_t _lseek(int fd, off_t offset, int whence);
    _read(int fd, void* buf, int n);
    _write(int fd, const void* buf, int n);
    _fsync(int fd);
    
    /*
     *	asynchronous file handler
     */
    static void create_filehandler(int fd, int mask, 
				   fileproc_t* proc, void* arg);
    static void block_filehandler(int fd);
    static void unblock_filehandler(int fd);
    static void destroy_filehandler(int fd, int* mask = 0, 
				    fileproc_t** proc = 0, void** arg = 0);
    static void destroy_all_filehandlers();
    static watching_filehandler(int fd);

    /*
     *        timer routines
     */
    static timerid_t create_timerhandler(int msec, timerproc_t proc, 
					 void* arg);
    static void destroy_timerhandler(timerid_t id);
    
    /*
     *	set synchronous and asynchronous flags
     */
    static void set_sync();
    static void set_async();

    /*
     *  cleaning up
     */
    static Thread poller_thread;
    static void poller(Thread* t);
    static void cleanup();
};

