/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: fixed.c,v 1.150 1996/09/05 19:23:43 nhall Exp $
 */
#define SM_SOURCE

#ifdef __GNUG__
#pragma implementation "fixed_s.h"
#pragma implementation "fixed_dir.h"
#pragma implementation "fixed.h"
#endif

#define FIXED_C
#include "sm_int.h"
#include "w_minmax.h"
#include "sm_du_stats.h"
#include "remote_client.h"

#include "bitmap.h"
#include "fixed.h"


MAKEPAGECODE(fixed_dir_p, page_p) ;  

void
fixed_dir_p::ntoh()
{
    /*
     *  BUGBUG -- fill in when appropriate
     */
    W_FATAL(eINTERNAL);
}

//
// format the directory page
//
rc_t
fixed_dir_p::format(const lpid_t& pid, tag_t tag, uint4_t flags)
{
    W_DO( page_p::format(pid, tag, flags, false) );
    return RCOK;
}

void
fixed_dir_p::set_full_bit(const int4 idx, const int4 offset)
{
    w_assert1( idx >= 0 && idx < nrecords() ); 
    w_assert1( offset >= 0 && offset < ss_m::ext_sz ); 

    bm_set((u_char *)bitmap_start_addr(), ss_m::ext_sz*idx+offset) ;    
}

rc_t
fixed_dir_p::insert_rec(int4& idx, shpid_t rec)
{
   if ( !is_space_available() )
     return RC(eRECWONTFIT) ; 

   // Find the slot 
   idx = nrecords() - nvacant() ;  

    // Insert the record    
   memcpy( record_addr(idx), (void *)&rec, sizeof(shpid_t) ) ; 

   // Reduce the number of vacant slots
   _pp->nvacant-- ; 

   return RCOK ; 
}
 
rc_t
fixed_dir_p::retrieve_rec(const int4 idx, shpid_t& rec)
{
   w_assert1( idx >= 0 && idx < nrecords() );

   //
   // Check whether the location contains a valid record
   // TO DO 
   // if ( !is_record_valid(idx) ) {
   //   return RC(eRECNOTEXISTS) ;
   // }

   // Retrieve the record 
   rec = *(shpid_t *)record_addr( idx ) ; 

   return RCOK ; 
}

rc_t
fixed_dir_p::retrieve_rec(const int4 idx, shpid_t*& rec)
{
   w_assert1( idx >= 0 && idx < nrecords() - nvacant() );

   // Check whether the location contains a valid record
   if ( !is_record_valid(idx) )
   {
      rec = NULL ;
      return RC(eRECNOTEXISTS) ;
   }

   // Retrieve the record
   rec = (shpid_t *)record_addr( idx ) ;

   return RCOK ;
} 

MAKEPAGECODE(fixed_p, page_p) ;  

void
fixed_p::ntoh()
{
    /*
     *  BUGBUG -- fill in when appropriate
     */
    W_FATAL(eINTERNAL);
}

//
// format the data page
//
rc_t
fixed_p::format(const lpid_t& pid, tag_t tag, uint4_t flags)
{
   W_DO( page_p::format(pid, tag, flags, false) );

   return RCOK;
}

/*********************************************************************
 *
 *  fixed_p::insert_rec(const int4 idx, const vec_t& data)
 *
 *  Insert the record (data) at the location 'idx'
 *
 *********************************************************************/
rc_t 
fixed_p::insert_rec(const slotid_t idx, const vec_t& data) 
{
   w_assert1( idx >= 0 && idx < nrecords() );
   w_assert1( data.size() >= record_size() && 
              data.size() <= aligned_record_size() ) ; 

   if ( idx >= nrecords() )
      return RC(eRECWONTFIT) ; 
  
   // Check whether the location is occupied 
   if ( is_record_valid(idx) )
      return RC(eRECEXISTS) ;  

   // Insert the record
   data.copy_to( record_addr(idx) ) ; 

   // Reduce the number of vacant slots
   _pp->nvacant-- ; 

   // Set the bitmap 
   bm_set( (u_char *)bitmap_start_addr(), idx ) ; 
   return RCOK ; 
}

/*********************************************************************
 *
 *  fixed_p::insert_rec(int4& idx, const vec_t& data)
 *
 *  Insert the record (data) at the first free location and return its
 *  index.
 *
 *********************************************************************/
rc_t 
fixed_p::insert_rec(const vec_t& data, slotid_t& idx) 
{
   w_assert1( data.size() == record_size() ) ;  

   if ( !is_space_available() )
     return RC(eRECWONTFIT) ; 

   // Get the first free record 
   idx = bm_first_set( (u_char *)bitmap_start_addr(), nrecords(), 0 ) ;  

    // Insert the record    
   data.copy_to( record_addr(idx) ) ; 

   // Reduce the number of vacant slots
   _pp->nvacant-- ; 

   // Set the bitmap 
   bm_set( (u_char *)bitmap_start_addr(), idx ) ;  
   return RCOK ; 
}

/*********************************************************************
 *
 *  fixed_p::update_rec(const int4 idx, smsize_t start, const vec_t& data)
 *
 *  Update the record (data) at the given location 'idx'  
 *  starting from the 'start' offset.
 *   
 *********************************************************************/
rc_t 
fixed_p::update_rec(const slotid_t idx, const smsize_t start, const vec_t& data)
{
   w_assert1( idx >= 0 && idx < nrecords() );
   w_assert1( data.size() <= (record_size() - start) ) ; 

   // Check whether the location contains a valid record 
   if ( !is_record_valid(idx) )
      return RC(eRECNOTEXISTS) ;  

    // update the record    
   data.copy_to( record_addr(idx) + start ) ; 
   return RCOK ; 
}

/*********************************************************************
 *
 *  fixed_p::retrieve_rec(const slotid_t idx, const vec_t& data)
 *
 *  Retrieve the record (data) at the given location 'idx'  
 *  from 'start' 
 *
 *********************************************************************/
rc_t
fixed_p::retrieve_rec(const slotid_t idx, const smsize_t start, uint4& len, void*& data)
{
    w_assert1( idx >= 0 && idx < nrecords() );
    w_assert1( smsize_t(start+len) <= record_size() ) ; 

    if ( smsize_t(start) >= record_size()) 
        return RC(eBADSTART);

    if ( smsize_t(start+len) >= record_size() )
        len = record_size() - start ; 
 
    // Check whether the location contains a valid record
    if ( !is_record_valid(idx) )
    {
       data = NULL ;     
       return RC(eRECNOTEXISTS) ;
    }

    // Retrieve the actual record 
    data = record_addr( idx ) + start; 
    return RCOK ; 
}

rc_t
fixed_p::retrieve_rec(const slotid_t idx, void*& data)
{
    w_assert1( idx >= 0 && idx < nrecords() );

    // Check whether the location contains a valid record
    if ( !is_record_valid(idx) )
    {
       data = NULL ;     
       return RC(eRECNOTEXISTS) ;
    }

    // Retrieve the actual record 
    data = record_addr( idx ) ; 
    return RCOK ; 
}

/*********************************************************************
 *
 *  fixed_p::delete_rec(const int4 idx)
 *
 *  Delete the record at the given location 
 *
 *********************************************************************/
rc_t                  
fixed_p::delete_rec(const slotid_t idx) 
{
   w_assert1( idx >= 0 && idx < nrecords() );

   // Update the bitmap
   bm_clr( (u_char *)bitmap_start_addr(), idx ) ; 
   return RCOK ; 
}

void
fixed_p::clear_bitmap()
{
   bm_zero( (u_char *)bitmap_start_addr(), mul8(bitmap_size()) - 1 ) ;
}
