/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: fixed_meta.h,v 1.85 1996/09/04 18:59:34 nhall Exp $
 */
#ifndef FIXED_META_H
#define FIXED_META_H

#ifdef __GNUG__
#pragma interface
#endif

/*--------------------------------------------------------------*
 *  class fixed_dir_p                                           *
 *                                                              * 
 * Fixed length record dir page class. It doesn't contain any   * 
 * slot, header and other record related space overhead         * 
 * associated with slotted pages.                               * 
 *                                                              *
 *      Initial Version 11/08/96 -- Karthikeyan Ramasamy        *
 *                                                              *
 *--------------------------------------------------------------*/

class fixed_dir_p : public page_p {
  
   public:

     // free space on fixed_dir_p is page_p 
     enum { data_sz = page_p::data_sz };
  
     MAKEPAGE(fixed_dir_p, page_p, 1);//Macro to install basic funcs from page_p 
     bool                  is_space_available() const ;

     int4                  data_start_offset() const ;
     void*                 data_start_addr() const ;

     int4                  bitmap_size() const ;
     int4                  bitmap_start_offset() const ;
     void*                 bitmap_start_addr() const ;
     void                  clear_bitmap() ;
     void                  set_full_bit(const int4 idx, const int4 offset) ;  

     static int4           max_dir_records() ; 
     int4                  nrecords() const ;
     int4                  nvacant() const ; 
     static smsize_t       record_size() ;
     void*                 record_addr(const int4 idx) const ;
     bool                  is_record_valid(const int4 idx) const ;

     bool                  is_fixed_dir_p() const ;

     rc_t                  insert_rec(int4& idx, shpid_t data) ;
     rc_t                  retrieve_rec(const int4 idx, shpid_t& data);
     rc_t                  retrieve_rec(const int4 idx, shpid_t*& data);
};

inline int4 fixed_dir_p::nvacant() const
{
   return _pp->nvacant ; 
} 

inline bool fixed_dir_p::is_space_available() const
{
   return nvacant() > 0  ;  
}

inline int4 fixed_dir_p::nrecords() const
{
   return page_p::nslots() ;
}

inline int4 fixed_dir_p::bitmap_size() const
{
   return data_sz - nrecords() * sizeof(shpid_t) ;  
}

inline int4 fixed_dir_p::bitmap_start_offset() const
{
   return page_p::hdr_size() + data_sz - bitmap_size() ;
}

inline void* fixed_dir_p::bitmap_start_addr() const
{
   return _pp->data + data_sz - bitmap_size() ;
}

inline int4 fixed_dir_p::data_start_offset() const
{
   // Sum of page header and header for fixed_dir_p
   return page_p::hdr_size() ;
}

inline void fixed_dir_p::clear_bitmap()
{
   bm_clr( (u_char *)bitmap_start_addr(), mul8(bitmap_size()) ) ;    
}

inline void* fixed_dir_p::data_start_addr() const
{
   return _pp->data ;
}

inline smsize_t fixed_dir_p::record_size() 
{
   return sizeof(shpid_t);  
}

inline void* fixed_dir_p::record_addr(const int4 idx) const
{
   return data_start_addr() + idx * record_size() ;
}

inline bool fixed_dir_p::is_record_valid(const int4 idx) const
{
   return idx >= 0 ? true : false ; // to keep the compiler happy 
}

inline bool fixed_dir_p::is_fixed_dir_p() const
{
   // all pages in file must be t_fixed
   w_assert3(tag()&t_fixed_dir_p);
   return tag()&t_fixed_dir_p;
}

inline int4 fixed_dir_p::max_dir_records()
{
   double tmp = fixed_dir_p::record_size() + div8(ss_m::ext_sz) + 1.0/8 ;

   return int4(double((fixed_dir_p::data_sz - 2)) / tmp) ;
}


#endif 
