/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: fixed.c,v 1.150 1996/09/05 19:23:43 nhall Exp $
 */
#define SM_SOURCE

#ifdef __GNUG__
#pragma implementation "fixed_m.h"
#endif

#define FIXED_C
#include "sm_int.h"
#include "w_minmax.h"
#include "sm_du_stats.h"
#include "remote_client.h"

#ifdef __GNUC__
template class w_list_t<blk_dir_t>;
#endif

/*********************************************************************
 *
 *  fixed_m::create(  stid_t     stid,
 *                    lpid_t&    first_page )
 *
 *  Create a fixed length file. 
 * 
 *********************************************************************/
rc_t
fixed_m::create( stid_t stid, lpid_t& first_page, int4 record_size )
{
   fixed_dir_p  page;

   W_DO( _alloc_data_extent( stid, record_size, lpid_t::eof, 
                             first_page) );
     
   return RCOK;
}
 
rc_t
fixed_m::create_rec( stid_t              fid,
                     const vec_t&        data,
                     sdesc_t&            sd,
                     int4&               idx )
{
   return RCOK ; 
}

rc_t 
fixed_m::create_rec( stid_t              fid,
                     const vec_t&        data,
                     sdesc_t&            sd,
                     rid_t&              rid )
{
   return RCOK ; 
}

rc_t
fixed_m::_purge( stid_t stid )
{
    lpid_t pid;
    W_DO( io->first_page(stid, pid) );
    while (pid.page)  {
        /*
         *  save current pid, get next pid, free current pid.
         */
        lpid_t cur = pid;
        rc_t rc = io->next_page(pid);
        W_DO( io->free_page(cur) );
        if (rc)  {
            if (rc.err_num() != eEOF)  {
                return RC_AUGMENT(rc);
            }
            break;
        }
    }

    return RCOK;
}

rc_t
fixed_m::_insert_blkld_entry( stid_t                 fid,
                                  lpid_t                 near,
                                  fixed_pid_t            ext_id,
                                  w_list_t<blk_dir_t>&   blkstack,
                                  bool                   last_page_full) 
{
   lpid_t        new_dir_ext ; 
   blk_dir_t*    stk_entry = NULL ; 
   int4          idx; 

   if ( blkstack.is_empty() ) 
   {
      // Allocate a new directory extent
      W_DO( _alloc_dir_extent( fid, near, new_dir_ext ) ) ;   

      // Allocate a new stack entry and initialize it
      stk_entry = new blk_dir_t ; 
      w_assert1( stk_entry ) ; 

      stk_entry->dir_fpid = fixed_pid_t(new_dir_ext.page, 0) ; 
 
      // Bring the page and latch it until the bulk load is done
      lpid_t      new_pid( fid, stk_entry->dir_fpid.pid() ) ; 
      W_DO( stk_entry->dir_page.fix( new_pid , LATCH_EX ) );  

      blkstack.push( stk_entry ) ; 
   } 
   
   stk_entry = blkstack.pop() ;  

   // Try inserting the record 
   rc_t rc = stk_entry->dir_page.insert_rec(idx, ext_id.ext_start()) ;  

   if ( rc == RC(eRECWONTFIT) ) 
   {
      // If record does not fit, move to the next page in the extent
      stk_entry->dir_fpid++ ; 
     
      // If the extent is full
      if ( stk_entry->dir_fpid.offset() >= ss_m::ext_sz ) 
      {    
         // Insert the current extent to its parent 
         W_DO( _insert_blkld_entry( fid, near, stk_entry->dir_fpid, 
                                        blkstack ) ) ; 

         // Allocate a new directory extent
         W_DO( _alloc_dir_extent( fid, near, new_dir_ext ) ) ;   

         // Initialize the stack entry 
         stk_entry->dir_fpid = fixed_pid_t(new_dir_ext.page,0) ; 
      } 

      // Bring the page and latch it until the bulk load is done
      lpid_t     new_pid( fid, stk_entry->dir_fpid.pid() ) ; 
      W_DO( stk_entry->dir_page.fix( new_pid, LATCH_EX) );  

      // Now insert the record
      W_DO( stk_entry->dir_page.insert_rec( idx, ext_id.ext_start() ) ) ; 
   }

   // Turn the bits that keep track of the free space off 
   for ( int i = 0 ; i < ext_id.offset() ; i++ )
   {
      stk_entry->dir_page.set_full_bit( idx, i ) ;       
   } 

   if ( last_page_full )
      stk_entry->dir_page.set_full_bit( idx, ext_id.offset() ) ;       

   // Inform the buffer pool that this page has been modified
   stk_entry->dir_page.set_dirty() ;  

   blkstack.push( stk_entry ) ; 
   return RCOK ; 
}

rc_t
fixed_m::_blkld_insert_rec( stid_t                 fid, 
                            int4                   record_size,
                            fixed_p*               curr_page, 
                            int4&                  curr_rec, 
                            fixed_pid_t&           curr_fpid,
                            const vec_t&           data,
                            w_list_t<blk_dir_t>&   blkld_stk )
{
    // Check for space in the current page 
    if ( !curr_page[curr_fpid.offset()].is_space_available() ) 
    {
       // Move on to the next page in the current extent 
       curr_fpid++ ; 

       // Check if data extent is full
       if ( curr_fpid.offset() >= ss_m::ext_sz ) 
       {
          lpid_t new_data_ext ; 
 
          curr_fpid-- ; 

          // Insert this data extent into the parent dir extent
          W_DO( _insert_blkld_entry( fid, lpid_t::eof, curr_fpid, blkld_stk) ) ;  

          // Allocate a new data extent
          W_DO( _alloc_data_extent( fid, record_size, 
                                    lpid_t::eof, new_data_ext ) ) ; 

          // Make all the pages in the current extent point to 
          // the next extent
          for ( int i = 0 ; i < ss_m::ext_sz ; i++ )
          {
              W_DO( curr_page[i].link_extent( new_data_ext.page ) ) ; 
          }

          // Link the last page of the current ext and the first 
          // page of the new ext. 
          W_DO( curr_page[curr_fpid.offset()].link_up( 
                                 curr_page[curr_fpid.offset()].prev(), 
                                 new_data_ext.page ) ); 

          // Unfix all the pages in the current extent
          for ( int i = 0 ; i < ss_m::ext_sz ; i++ )
          {
              curr_page[i].unfix() ; 
          }

          curr_fpid = fixed_pid_t( new_data_ext.page, 0 ) ; 
       }

       lpid_t  new_data_pid( fid, curr_fpid.pid() ) ;  
       W_DO( curr_page[curr_fpid.offset()].fix(new_data_pid, LATCH_EX) ) ; 
       curr_rec = 0 ; 
    }
               
    W_DO(curr_page[curr_fpid.offset()].insert_rec((const int)curr_rec, data)) ;  
    curr_page[curr_fpid.offset()].set_dirty() ; 
    curr_rec++ ; 

    return RCOK ; 
}

rc_t
fixed_m::bulk_load(  stid_t              fid,
                     stid_t              source,
                     sdesc_t&            sd )
{
   // Purge the entire fixed length file 
   W_DO( _purge(fid) ) ; 

   w_list_t<blk_dir_t>    blkld_stk(offsetof(blk_dir_t, link)) ; 
   fixed_p                curr_page[ss_m::ext_sz] ;  
   int4                   curr_rec = 0 ; 
   lpid_t                 curr_lpid ;
   int                    rec_count = 0 ; // For debugging purposes

   // Create the first data extent 
   W_DO( _alloc_data_extent( fid, sd.sinfo().record_size, lpid_t::eof, 
                             curr_lpid) );

   fixed_pid_t            curr_fpid(curr_lpid.page,0) ; 

   // Fix the first data page in the fact file
   W_DO( curr_page[curr_fpid.offset()].fix( curr_lpid, LATCH_EX ) );  

   /*
    *  Go thru the src file page by page
    */
   int i = 0;          // toggle
   file_p page[2];     // page[i] is current page

   rc_t rc;
   const record_t* pr = 0;     // previous record
   lpid_t pid;
   bool eof = false;
   bool skip_last = false;

   for (rc = fi->first_page(source, pid); !rc.is_error() && !eof;
        rc = fi->next_page(pid, eof))     
   {
        /*
         *  for each page ...
         */
        W_DO( page[i].fix(pid, LATCH_SH) );

        int s = page[i].next_slot(0);
        if (! s)  {
            /*
             *  do nothing. skip over empty page, so do not toggle
             */
            continue;
        }
        for ( ; s; s = page[i].next_slot(s))  
        {
            /*
             *  for each slot in page ...
             */
            record_t* r;
            W_COERCE( page[i].get_rec(s, r) );
            rec_count++ ; 

            if (pr) 
            {
                if ( pr->body_size() > sd.sinfo().record_size )
                {
                   cerr << "Truncating the size of the record " 
                        << rec_count << endl ;
                }

                vec_t key(pr->hdr(), pr->hdr_size());
                vec_t el(pr->body(), (int)pr->body_size());

                W_DO( _blkld_insert_rec( fid, sd.sinfo().record_size, 
                                         curr_page, curr_rec, curr_fpid, 
                                         el, blkld_stk ) ) ; 
            }
      
            if (page[1-i])  page[1-i].unfix();
            pr = r;

            if (!s) break;
        }
        i = 1 - i;      // toggle i
        if (eof) break;
    }

    if (rc)  {
        return rc.reset();
    }

    if (!skip_last && pr) 
    {
        vec_t key(pr->hdr(), pr->hdr_size());
        vec_t el(pr->body(), (int)pr->body_size());

        W_DO( _blkld_insert_rec( fid, sd.sinfo().record_size, curr_page, 
                                 curr_rec, curr_fpid, el, blkld_stk ) ) ; 
    }

    // Null Terminate the chain of pages
    W_DO( curr_page[curr_fpid.offset()].link_up(
                           curr_page[curr_fpid.offset()].prev(), 0) ) ; 
 
    for ( int j = 0 ; j <= curr_fpid.offset() ; j++ )
    {
        W_DO( curr_page[j].link_extent(0) ) ;
    }
       
    int4  tree_level = 0 ; 
    uint2 bitmap = 0 ; 

    // Code looks ugly... A better way to avoid repetition ?
    if ( !blkld_stk.is_empty() )  
    {
        // Insert the last data extent
        W_DO( _insert_blkld_entry( fid, lpid_t::eof, curr_fpid, blkld_stk,
               curr_page[curr_fpid.offset()].is_space_available() ? 
                                  false : true ) ) ;  

        blk_dir_t*   stk_entry = NULL ; 
        while ( blkld_stk.num_members() > 1 )
        {
           stk_entry = blkld_stk.pop() ; tree_level++ ;  

            W_DO( _insert_blkld_entry( fid, lpid_t::eof, stk_entry->dir_fpid, 
                      blkld_stk, 
                      stk_entry->dir_page.is_space_available() ? false : true ) ) ;  

           stk_entry->dir_page.unfix() ;  delete stk_entry ; 
        }

        // Build the top level bitmap for the last entry in the stack  
        stk_entry = blkld_stk.pop() ;  tree_level++ ; 
        for ( int j = 0 ; j < stk_entry->dir_fpid.offset() ; j++ )
        {
            bm_set((u_char *)&bitmap, j) ; 
        }
  
        // Check the last page 
        if ( !stk_entry->dir_page.is_space_available() ) 
            bm_set((u_char *)&bitmap, stk_entry->dir_fpid.offset()) ; 

        // Set the root page id in the directory
        sd.set_root( stk_entry->dir_fpid.ext_start() ) ;  

        stk_entry->dir_page.unfix() ;  delete stk_entry ; 
    }
    else
    {
        // Build the only top level 
        for ( int j = 0 ; j < curr_fpid.offset() ; j++ )
        {
            bm_set((u_char *)&bitmap, j) ; 
        }
  
        // Check the last page 
        if ( !curr_page[curr_fpid.offset()].is_space_available() ) 
            bm_set( (u_char *)&bitmap, curr_fpid.offset() ) ; 
   
        // Set the root page id in the directory
        sd.set_root( curr_fpid.ext_start() ) ;  
    }

    sd.set_tree_level(tree_level) ; 
    sd.set_free_bitmap(bitmap) ; 

    return RCOK ; 
}  

rc_t fixed_m::first_page(  const lpid_t&       root,
                      lpid_t&             pid,
                      const int4          record_size,
                      const int4          tree_level,
                      bool                lock,
                      pid_cache_t*        cache )
{
   slotid_t data_idx ;

   W_DO( get_data_page(root, 0, record_size, tree_level, pid, data_idx) ) ;

   return RCOK ;
}


/*
rc_t 
fixed_m::first_page(  const lpid_t&       root,
                      lpid_t&             pid,
                      const int4          record_size, 
                      const int4          tree_level,
                      bool                lock, 
                      pid_cache_t*        cache )
{
   int4  rec_idx ;

   fixed_pid_t   root_ext( root.page, 0 ) ;
   fixed_pid_t   first_fpid ; 
   
   W_DO ( _lookup( root.stid(), 0, record_size, root_ext, rec_idx, 
                   first_fpid, tree_level ) ) ; 

   pid = lpid_t( root.stid(), first_fpid.pid() ) ; 
 
   return RCOK ; 
}
*/

rc_t
fixed_m::last_page( const lpid_t&       root,
                     lpid_t&             pid,
                     const int4          record_size,
                     const int4          tree_level,
                     bool                lock,
                     pid_cache_t*        cache )
{
   return RCOK ; 
}

rc_t
fixed_m::next_page(  const lpid_t&       curr_pid,
                     lpid_t&             pid,
                     bool&               eof,
                     bool                lock,
                     pid_cache_t*        cache ) 
{
   fixed_p data_page ; 

   W_DO ( data_page.fix( curr_pid, LATCH_SH) ) ; 

   eof = data_page.next() ? false : true ; 
   pid = data_page.next() ? lpid_t(curr_pid.stid(), data_page.next()) 
                          : lpid_t::null ;  

   return RCOK ; 
}

int4
fixed_m::_records_per_dir_entry( const int4 record_size, int2 level )
{
   int4 result = fixed_p::max_data_records(record_size) << ss_m::ext_bits;    

   while ( level > 0 )
   {
       result *= fixed_dir_p::max_dir_records() << ss_m::ext_bits ; 
       level-- ; 
   }

   return result ; 
}

void
fixed_m::_calculate_idx( const int4   idx, 
                         const int4   record_size,
                         const int4   level,
                         int4&        new_idx,
                         int4&        pid_idx,
                         int4&        rec_idx )
{
    int4 num_records = _records_per_dir_entry(record_size, level-1)  ;

    pid_idx = idx / num_records / fixed_dir_p::max_dir_records() ;

    rec_idx = (idx / num_records) % fixed_dir_p::max_dir_records() ;

    // Calculate the idx for the next lower level
    new_idx = idx - (rec_idx + pid_idx * fixed_dir_p::max_dir_records()) * 
                               num_records ;
}

rc_t
fixed_m::_lookup( const stid_t        fid, 
                  const int4          idx,
                  const int4          record_size, 
                  const fixed_pid_t&  root_ext,
                  slotid_t&           data_idx,
                  fixed_pid_t&        data_fpid,
                  int                 clevel )
{
   int4 new_idx, rec_idx, page_idx ;

   if ( !clevel )
   {
       data_fpid = fixed_pid_t( root_ext.ext_start(),  
                     idx / fixed_p::max_data_records(record_size) ); 
       data_idx =  slotid_t(idx % fixed_p::max_data_records(record_size)) ; 
       return RCOK ;
   }

   _calculate_idx( idx, record_size, clevel, new_idx, page_idx, rec_idx ) ;

   //
   // Bring the page in
   //
   fixed_dir_p    dir_page ;
   fixed_pid_t    dir_fpid( root_ext.ext_start(), page_idx ) ; 
   lpid_t         dir_lpid( fid, dir_fpid.pid() ) ;

   W_DO( dir_page.fix( dir_lpid, LATCH_SH ) ) ;

   shpid_t        data_shpid ;
   W_DO( dir_page.retrieve_rec( rec_idx, data_shpid ) ) ;

   fixed_pid_t    new_fpid( data_shpid, 0 ) ; 
 
   W_DO( _lookup( fid, new_idx, record_size, new_fpid, data_idx, 
                  data_fpid, clevel-1 ) ) ;

   return RCOK ;
}

/*********************************************************************
 *
 *  fixed_m::read_rec(const int4 idx, 
 *                    int4       start,
 *                    int4&      len, 
 *                    void*      buf);
 *
 *  Read the record from the fixed length record file given its idx
 *  and starting offset in the record. 
 *
 *********************************************************************/
rc_t 
fixed_m::read_rec(
              const lpid_t&       root,
              const int4          idx,
              const int4          record_size,
              const int4          tree_level,  
              const smsize_t      start,
              uint4&              len,
              void*               buf )
{
    slotid_t      data_idx ;
    fixed_pid_t   data_fpid ;
    fixed_pid_t   root_fpid( root.page, 0 ) ; 

    W_DO( _lookup( root.stid(), idx, record_size, 
                   root_fpid, data_idx, data_fpid, tree_level) ) ;

    fixed_p data_page ;
    lpid_t  data_lpid( root.stid(), data_fpid.pid()) ; 

    W_DO( data_page.fix( data_lpid, LATCH_SH ) ) ;

    void*  rec ;
    W_DO( data_page.retrieve_rec( data_idx, start, len, rec ) ) ;
    memcpy( buf, rec, len ) ;

    return RCOK ;
}

rc_t 
fixed_m::read_rec(
              const rid_t&        s_rid,
              const smsize_t      start,
              uint4&              len,
              void*               buf )
{
    rid_t   rid(s_rid);
    fixed_p page;

    W_DO( _locate_page(rid, page, LATCH_SH) );

    if ( smsize_t(start) >= page.record_size()) {
        return RC(eBADSTART);
    }

    if ( smsize_t(start + len) >= page.record_size()) {
        len = page.record_size() - start;
    } 

    void* rec ; 
    W_DO ( page.retrieve_rec( rid.slot, start, len, rec ) ) ; 

    memcpy(buf, rec, len);
    return RCOK;
}

rc_t
fixed_m::update_rec(
             const lpid_t&        root,
             const int4           idx,
             const int4           record_size,
             const int4           tree_level,  
             const smsize_t       start,
             const vec_t&         data ) 
{ 
    slotid_t      data_idx ;
    fixed_pid_t   data_fpid ;
    fixed_pid_t   root_fpid( root.page, 0 ) ; 

    W_DO( _lookup( root.stid(), idx, record_size, 
                   root_fpid, data_idx, data_fpid, tree_level) ) ;

    fixed_p data_page ;
    lpid_t  data_lpid( root.stid(), data_fpid.pid()) ; 

    W_DO( data_page.fix( data_lpid, LATCH_EX ) ) ;

    W_DO( data_page.update_rec( data_idx, start, data ) ) ;
    data_page.set_dirty() ; 
 
    return RCOK ;
}

rc_t 
fixed_m::update_rec(
              const rid_t&        rid,
              const smsize_t      start,
              const vec_t&        data )
{
    fixed_p    page;
    void*      rec;

    W_DO(_locate_page(rid, page, LATCH_EX));

    /*
     *  Do some parameter checking
     */
    if (smsize_t(start) >= page.record_size()) {
        return RC(eBADSTART);
    }

    if (data.size() > (page.record_size()-start)) {
        return RC(eRECUPDATESIZE);
    }

    W_DO( page.update_rec( rid.slot, start, data ) ) ; 
    page.set_dirty() ; 

    return RCOK;
}

rc_t
fixed_m::destroy_rec( const lpid_t&       root,
                      const int4          idx,
                      const int4          record_size,
                      const int4          tree_level ) 
{
    return RCOK ; 
}
 
rc_t
fixed_m::destroy_rec( const rid_t& rid )
{
    fixed_p    page;
    void*      rec;

    W_DO( _locate_page(rid, page, LATCH_EX) );
    W_DO( page.delete_rec(rid.slot) ) ; 

    return RCOK ; 
}

rc_t
fixed_m::_alloc_dir_extent( stid_t             fid,
                            const lpid_t       near,
                            lpid_t&            alloc_eid)                      
{
   W_DO( _alloc_extent( fid, near, alloc_eid ) ) ; 
   for ( int i = ss_m::ext_sz - 1 ; i >= 0 ; i-- ) 
   {
       W_DO( _initialize_dir_page( fid, alloc_eid.page + i ) ) ;  
   }    

   return RCOK ; 
}

rc_t
fixed_m::_initialize_dir_page( stid_t          fid,
                               const shpid_t   pid )
{
   lpid_t            dir_pid(fid, pid) ; 
   fixed_dir_p       dir_page ;

   W_DO( dir_page.fix( dir_pid, LATCH_EX, dir_page.t_virgin) );

   dir_page.persistent_part().nslots = dir_page.max_dir_records() ;  
   dir_page.persistent_part().nvacant = dir_page.max_dir_records() ; 
   dir_page.clear_bitmap() ; 
   dir_page.set_dirty() ; 

   return RCOK ; 
}  

rc_t
fixed_m::_alloc_data_extent( stid_t             fid,
                             int4               record_size, 
                             const lpid_t       near,
                             lpid_t&            alloc_eid)
{
   fixed_p           page ; 

   // Allocate an extent
   W_DO( _alloc_extent( fid, near, alloc_eid ) ) ; 

   // Initialize the last page in the extent 
   W_DO( _initialize_data_page( fid, alloc_eid.page+ss_m::ext_sz - 1, 
                                alloc_eid.page+ss_m::ext_sz - 2,
                                lpid_t::null.page, record_size ) ) ;  

   // Iterate thro' the pages and initialize the pages' 
   // meta structures 
   for ( int i = ss_m::ext_sz - 2 ; i >= 1 ; i-- ) 
   {
       W_DO( _initialize_data_page( fid, alloc_eid.page+i, alloc_eid.page+i-1,
                                    alloc_eid.page+i+1, record_size ) ) ; 
   }    

   // Initialize the first page in the extent
   W_DO( _initialize_data_page( fid, alloc_eid.page, lpid_t::null.page,
                                alloc_eid.page+1, record_size ) ) ;  
   return RCOK ; 
}

rc_t
fixed_m::_initialize_data_page( stid_t          fid, 
                                const shpid_t   pid,
                                const shpid_t   prev_pid,
                                const shpid_t   next_pid, 
                                int4            record_size )
{
   lpid_t          data_pid( fid, pid ) ; 
   fixed_p         data_page ; 
   fixed_p_hdr_t   hdr ; 
  
   hdr.actual_record_size = record_size ; 

   W_DO( data_page.fix(data_pid, LATCH_EX, data_page.t_virgin) ) ;

   W_DO( data_page.write_header(hdr) ) ; 

   data_page.persistent_part().nslots = 
                             data_page.max_data_records(record_size) ;  

   data_page.persistent_part().nvacant = 
                             data_page.max_data_records(record_size) ; 
   data_page.clear_bitmap() ;  

   W_DO( data_page.link_up( prev_pid, next_pid ) ) ;  
   data_page.set_dirty() ; 

   return RCOK ; 
}

rc_t
fixed_m::_alloc_extent( stid_t             fid,
                        const lpid_t       near,
                        lpid_t&            alloc_eid )
{
   lpid_t               pids[ss_m::ext_sz];

   W_DO( io->alloc_pages(fid, near, ss_m::ext_sz, pids) );
   alloc_eid = pids[0] ; 

   return RCOK ; 
}

rc_t
fixed_m::_locate_page( const rid_t&        rid, 
                       fixed_p&            page, 
                       latch_mode_t        mode)
{
    DBG( <<"fixed_m::_locate_page rid=" << rid );

    /*
     * pin the page unless it's remote; even if it's remote,
     * pin if we are using page-level concurrency
     */
    if (cc_alg == t_cc_page || (! rid.pid.is_remote())) {
        W_DO(page.fix(rid.pid, mode));
    } else {
        W_FATAL(eINTERNAL);
    }

    w_assert3(page.pid().page == rid.pid.page);
    // make sure page belongs to pid.stid
    if (page.pid() != rid.pid) {
        page.unfix();
        return RC(eBADPID);
    }

    if (rid.slot < 0 || rid.slot >= page.nrecords())  {
        return RC(eBADSLOTNUMBER);
    }

    return RCOK;
}
