/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: fixed_m.h,v 1.85 1996/09/04 18:59:34 nhall Exp $
 */
#ifndef FIXED_M_H
#define FIXED_M_H

#ifdef __GNUG__
#pragma interface
#endif

#include "fixed.h"

struct blk_dir_t {
    fixed_pid_t   dir_fpid  ;
    fixed_dir_p   dir_page ;
    w_link_t      link ;
} ;

class fixed_m : public smlevel_2 {

    public:

      fixed_m() { } 
      ~fixed_m() { } 

      static rc_t create(
                      stid_t              stid, 
                      lpid_t&             first_page,
                      int4                record_size) ;

      static rc_t create_rec( 
                      stid_t              fid, 
                      const vec_t&        data, 
                      sdesc_t&            sd, 
                      int4&               idx);

      static rc_t create_rec( 
                      stid_t              fid, 
                      const vec_t&        data, 
                      sdesc_t&            sd, 
                      rid_t&              rid);

      static rc_t insert_rec( 
                      stid_t              fid, 
                      const vec_t&        data, 
                      sdesc_t&            sd, 
                      const int4          idx);

      // 
      // Read the given record. 'start' can be used to retrieve the
      // attribute directly 
      //
      static rc_t read_rec(
                      const lpid_t&       root,
                      const int4          idx, 
                      const int4          record_size,
                      const int4          tree_level, 
                      const smsize_t      start, 
                      uint4&              len, 
                      void*               buf);

      static rc_t get_data_page(
                      const lpid_t&       root,
                      const int4          idx,
                      const int4          record_size,
                      const int4          tree_level,
                      lpid_t&             data_page,
                      slotid_t&           data_idx ) ; 

      static rc_t get_data_page(
                      const lpid_t&       root,
                      const int4          idx,
                      const int4          record_size,
                      const int4          tree_level,
                      lpid_t&             data_page,
                      slotid_t&           data_idx,
                      fixed_pid_t&        data_fpid ) ; 

      static rc_t read_rec(
                      const rid_t&        rid, 
                      const smsize_t      start, 
                      uint4&              len, 
                      void*               buf);

      static rc_t read_rec(
                      const lpid_t&       root,
                      const int4          idx, 
                      const int4          record_size,
                      const int4          tree_level, 
                      uint4&              len, 
                      void*               buf) ;

      static rc_t read_rec(
                      const rid_t&        rid, 
                      uint4&              len, 
                      void*               buf) ;

      //
      //  Modify the existing records
      //
      static rc_t update_rec(
                      const rid_t&        rid, 
                      const smsize_t      start, 
                      const vec_t&        data);

      static rc_t update_rec(
                      const lpid_t&       root,
                      const int4          idx, 
                      const int4          record_size,
                      const int4          tree_level, 
                      const smsize_t      start, 
                      const vec_t&        data);

      static rc_t update_rec( 
                      const rid_t&        rid, 
                      const vec_t&        data);

      static rc_t update_rec( 
                      const lpid_t&       root,
                      const int4          idx, 
                      const int4          record_size,
                      const int4          tree_level, 
                      const vec_t&        data);

      //
      // Delete the existing records
      //
      static rc_t destroy_rec(
                      const lpid_t&       root,
                      const int4          idx,
                      const int4          record_size,
                      const int4          tree_level ) ; 
 
      static rc_t destroy_rec( const rid_t& rid );

      static rc_t bulk_load( 
                     stid_t              fid,
                     stid_t              source,
                     sdesc_t&            sd ) ;

      static rc_t first_page(
                     const lpid_t&       root,
                     lpid_t&             pid,
                     const int4          record_size,
                     const int4          tree_level,
                     bool                lock = FALSE,
                     pid_cache_t*        cache = NULL);

      static rc_t next_page(
                     const lpid_t&       curr_pid,
                     lpid_t&             pid,
                     bool&               eof,
                     bool                lock = FALSE,
                     pid_cache_t*        cache = NULL);

      static rc_t last_page(
                     const lpid_t&       root,
                     lpid_t&             pid,
                     const int4          record_size,
                     const int4          tree_level,
                     bool                lock = FALSE,
                     pid_cache_t*        cache = NULL);

      static rc_t locate_page(const rid_t& rid, fixed_p& page,
                           latch_mode_t mode)
                           {return _locate_page(rid, page, mode); }
 
    protected :

      static rc_t _purge(stid_t          fid) ; 

      static rc_t _locate_page( 
                     const rid_t&         rid, 
                     fixed_p&             page, 
                     latch_mode_t         mode );

      static rc_t _insert_blkld_entry(
                     stid_t               fid,
                     lpid_t               near, 
                     fixed_pid_t          ext_id,
                     w_list_t<blk_dir_t>& blkstack,
                     bool                 last_page_full = true ) ; 

      static rc_t _alloc_dir_extent(
                     stid_t               fid,
                     const lpid_t         near,
                     lpid_t&              alloc_eid) ;

      static rc_t _alloc_data_extent( 
                     stid_t               fid,
                     int4                 record_size,
                     const lpid_t         near,
                     lpid_t&              alloc_eid) ;

      static rc_t _alloc_extent(
                    stid_t                fid,
                    const lpid_t          near,
                    lpid_t&               alloc_eid) ; 

      static rc_t _initialize_dir_page( 
                    stid_t                fid,
                    const shpid_t         pid) ;

      static rc_t _initialize_data_page( 
                    stid_t                fid,
                    const shpid_t         pid,
                    const shpid_t         prev_pid,
                    const shpid_t         next_pid,
                    int4                  record_size ) ;

      static rc_t _blkld_insert_rec(
                    stid_t                fid,
                    int4                  record_size,
                    fixed_p*              curr_page,
                    int4&                 curr_rec,
                    fixed_pid_t&          curr_fpid,
                    const vec_t&          data,  
                    w_list_t<blk_dir_t>&  blkld_stk ) ; 
 
      static void _calculate_idx( 
                    const int4            idx,
                    const int4            record_size,
                    const int4            level,
                    int4&                 new_idx,
                    int4&                 pid_idx,
                    int4&                 record_idx ) ;

      static rc_t _lookup( 
                    const stid_t          fid,
                    const int4            idx,
                    const int4            record_size,
                    const fixed_pid_t&    root_ext,
                    slotid_t&             data_idx,
                    fixed_pid_t&          data_fpid,
                    int                   clevel ) ;

      static int4 _records_per_dir_entry(
                    const int4            record_size,
                    int2                  level ) ;   

    private:

       // disabled
       fixed_m(const fixed_m&);
       operator=(const fixed_m&);
} ;

inline rc_t fixed_m::read_rec( const lpid_t&      root,
                               const int4         idx,
                               const int4         record_size, 
                               const int4         tree_level, 
                               uint4&             len,
                               void*              buf)
{
    W_DO( read_rec(root, idx, record_size, tree_level, 0, len, buf) ) ;
    return RCOK ; 
}

inline rc_t fixed_m::read_rec( const rid_t&       rid,
                               uint4&             len,
                               void*              buf)
{
    W_DO( read_rec(rid, 0, len, buf) ) ;
    return RCOK ; 
}

inline rc_t fixed_m::update_rec( const lpid_t&      root,
                                 const int4         idx,
                                 const int4         record_size, 
                                 const int4         tree_level, 
                                 const vec_t&       data)
{
    W_DO( update_rec(root, idx, record_size, tree_level, 0, data) ) ;
    return RCOK ; 
}

inline rc_t fixed_m::update_rec( const rid_t&        rid,
                                 const vec_t&        data)
{
    W_DO( update_rec(rid, 0, data) ) ;
    return RCOK ; 
}

inline rc_t fixed_m::get_data_page( const lpid_t&    root,
                                    const int4       idx,
                                    const int4       record_size,
                                    const int4       tree_level,
                                    lpid_t&          data_page,
                                    slotid_t&        data_idx )
{
   fixed_pid_t   data_fpid ;
   fixed_pid_t   root_fpid( root.page, 0 ) ;

   W_DO( _lookup( root.stid(), idx, record_size,
                  root_fpid, data_idx, data_fpid, tree_level) ) ; 
   data_page = lpid_t( root.stid(), data_fpid.pid() ) ; 
   return RCOK ; 
}

inline rc_t fixed_m::get_data_page( const lpid_t&    root,
                                    const int4       idx,
                                    const int4       record_size,
                                    const int4       tree_level,
                                    lpid_t&          data_page,
                                    slotid_t&        data_idx,
                                    fixed_pid_t&     data_fpid )
{
   fixed_pid_t   root_fpid( root.page, 0 ) ;

   W_DO( _lookup( root.stid(), idx, record_size,
                  root_fpid, data_idx, data_fpid, tree_level) ) ; 
   data_page = lpid_t( root.stid(), data_fpid.pid() ) ; 
   return RCOK ; 
}

#endif 
