/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: fixed_s.h,v 1.23 1996/08/16 15:23:29 nhall Exp $
 */

#ifndef FIXED_S_H
#define FIXED_S_H

#ifdef __GNUG__
#pragma interface
#endif

/*--------------------------------------------------------------*
 *  class fixed_p_hdr_t                                         *
 *                                                              *
 *  This is the header specific to a fixed length record page.  * 
 *  It is stored as the first record in the data area of the    *
 *  page.                                                       *
 *                                                              *
 *      Initial Version 11/08/96 -- Karthikeyan Ramasamy        *
 *                                                              *
 *--------------------------------------------------------------*/

struct fixed_p_hdr_t {
    uint4    actual_record_size;
    shpid_t  next_ext ; 
};

struct fixed_pid_t {    
    shpid_t    _extent_start ;
    int4       _offset ; 

    fixed_pid_t();
    fixed_pid_t(const shpid_t& s, int4 p);

    shpid_t   ext_start()   const {return _extent_start;}
    int4      offset() const {return _offset;}
    shpid_t   pid() const {return _extent_start+_offset;} 

    bool operator==(const fixed_pid_t& p) const;
    bool operator!=(const fixed_pid_t& p) const;
    fixed_pid_t& operator++() ;
    fixed_pid_t& operator++(int) ;
    fixed_pid_t& operator--() ;
    fixed_pid_t& operator--(int) ;
    friend ostream& operator<<(ostream&, const fixed_pid_t& p);

    static const fixed_pid_t bof;
    static const fixed_pid_t eof;
    static const fixed_pid_t null;
} ;

inline fixed_pid_t::fixed_pid_t() : _offset(-1) { }

inline fixed_pid_t::fixed_pid_t(const shpid_t& s, int4 p) : _extent_start(s), _offset(p)
{}

inline bool fixed_pid_t::operator==(const fixed_pid_t& p) const
{
    return (_extent_start == p._extent_start) && (_offset == p._offset);
}

inline bool fixed_pid_t::operator!=(const fixed_pid_t& p) const
{
    return !(*this == p);
}

inline fixed_pid_t& fixed_pid_t::operator++()
{
    ++_offset ; return *this ;  
}

inline fixed_pid_t& fixed_pid_t::operator++(int)
{
    _offset++ ; return *this ;  
}

inline fixed_pid_t& fixed_pid_t::operator--()
{
    --_offset ; return *this ;  
}

inline fixed_pid_t& fixed_pid_t::operator--(int)
{
    _offset-- ; return *this ;  
}

inline ostream& operator<<(ostream& o, const fixed_pid_t& pid)
{
    return o << "p(" << pid.ext_start() << '.' << pid.offset() << ')';
}


#endif 

