/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: plain_m.c,v 1.150 1996/09/05 19:23:43 nhall Exp $
 */
#define SM_SOURCE

#ifdef __GNUG__
#pragma implementation "plain_m.h"
#endif

#define PLAIN_C
#include "sm_int.h"
#include "w_minmax.h"
#include "sm_du_stats.h"
#include "remote_client.h"

/*********************************************************************
 *
 *  plain_m::create(  stid_t     stid,
 *                    lpid_t&    first_page )
 *
 *  Create a plain fixed length file. 
 * 
 *********************************************************************/
rc_t
plain_m::create( stid_t stid, lpid_t& first_page, int4 record_size )
{
   W_DO( _alloc_data_page( stid, record_size, lpid_t::eof, 
                             first_page) );
     
   return RCOK;
}

rc_t 
plain_m::_alloc_data_page( stid_t               fid,
                           int4                 record_size,
                           const lpid_t         near,
                           lpid_t&              pageid) 
{
   W_DO( io->alloc_new_last_page(fid, near, pageid) ) ; 

   fixed_p          data_page ; 
   fixed_p_hdr_t    hdr ;

   W_DO( data_page.fix( pageid, LATCH_EX, data_page.t_virgin) ) ; 

   hdr.actual_record_size = record_size ; 
   W_DO( data_page.write_header(hdr) ) ; 

   data_page.persistent_part().nslots = 
                        data_page.max_data_records(record_size) ; 

   data_page.persistent_part().nvacant = 
                        data_page.max_data_records(record_size) ; 

   data_page.clear_bitmap() ;

   W_DO( data_page.link_up( near.page, 0 ) ) ;  
   data_page.set_dirty() ; 

   return RCOK ;
}

rc_t
plain_m::append_rec( stid_t         fid,
                     sdesc_t&       sd,
                     const vec_t&   data,
                     fixed_p&       last_page,
                     rid_t&         rid )
{
   w_assert3(fid == sd.stpgid().stid());
 
   lpid_t pid(fid, sd.sinfo().last);

   if(last_page.is_fixed()) {
        w_assert3(last_page.pid() == pid);
   } else {
        W_DO( last_page.fix( pid, LATCH_EX ) ) ;
   }

   // Check space availability and insert
   if ( !last_page.is_space_available() )
   {
      lpid_t   new_pid ; 

      // Allocate a new page and link the previous last 
      // page
      W_DO( _alloc_data_page( fid, sd.sinfo().record_size, pid, new_pid ) ) ; 

      // Unfix the old page and fix the new page
      W_DO( last_page.fix( new_pid, LATCH_EX ) ) ;

      // Update the last page in the directory entry
      sd.set_last_page( new_pid.page ) ; 
   }

   slotid_t  idx ; 

   W_DO( last_page.insert_rec( data, idx) ) ; 

   rid.pid = last_page.pid() ; rid.slot = idx ; 

   last_page.set_dirty() ; 

   return RCOK ; 
}

rc_t
plain_m::next_page(  const lpid_t&       curr_pid,
                     lpid_t&             pid,
                     bool&               eof,
                     bool                lock,
                     pid_cache_t*        cache )
{
   W_DO( fixed_m::next_page( curr_pid, pid, eof, lock, cache ) ) ; 

   return RCOK ; 
}
