/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: smfile.c,v 1.29 1996/10/28 21:45:36 nhall Exp $
 */
#define SM_SOURCE
#define SMFIXED_C

#include "w.h"
#include "option.h"
#include "sm_int.h"
#include "device.h"
#include "vol.h"
#include "app_support.h"
#include "remote_client.h"

#define DBGTHRD(arg) DBG(<<" th."<<me()->id<<" tid."<<xct()->tid()<<" " arg)

/*==============================================================*
 *  Physical ID version of all the storage operations           *
 *==============================================================*/

/*--------------------------------------------------------------*
 *  ss_m::create_fixed_file()				        *
 *--------------------------------------------------------------*/
rc_t
ss_m::create_fixed_file(
    vid_t 			vid, 
    stid_t& 			fid, 
    store_property_t 		property,
    uint4  		        length)
{
    SM_PROLOGUE_RC(ss_m::create_fixed_file, in_xct, 0);
    RES_SMSCRIPT(<<"create_fixed_file " <<vid << " " << property << " " << length );
    W_DO(_create_fixed_file(vid, fid, property, length));
    DBGTHRD(<<"create_fixed_file returns " << fid);
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::destroy_fixed_file()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::destroy_fixed_file(const stid_t& fid)
{
    SM_PROLOGUE_RC(ss_m::destroy_fixed_file, in_xct, 0);
    RES_SMSCRIPT(<<"destroy_fixed_file " <<fid);
    W_DO(_destroy_fixed_file(fid));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::create_fixed_rec()			                *
 *--------------------------------------------------------------*/
rc_t
ss_m::create_fixed_rec(const stid_t& fid, const vec_t& data, 
                       rid_t& new_rid)
{
    SM_PROLOGUE_RC(ss_m::create_rec, in_xct, 0);
    RES_SMSCRIPT(<<"create_rec " <<fid
	<<" " << data ) ; 
    W_DO(_create_fixed_rec(fid, data, new_rid));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::create_fixed_rec()			                *
 *--------------------------------------------------------------*/
rc_t
ss_m::create_fixed_rec(const stid_t& fid, const vec_t& data, 
                       uint4& idx)
{
    SM_PROLOGUE_RC(ss_m::create_rec, in_xct, 0);
    RES_SMSCRIPT(<<"create_rec " <<fid
	<<" " << data ) ; 
    W_DO(_create_fixed_rec(fid, data, idx));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::destroy_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::destroy_fixed_rec(const rid_t& rid)
{
    SM_PROLOGUE_RC(ss_m::destroy_fixed_rec, in_xct, 0);
    RES_SMSCRIPT(<<"destroy_fixed_rec " <<rid);
    W_DO(_destroy_fixed_rec(rid));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::destroy_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::destroy_fixed_rec(const stid_t& fid, const uint4 idx)
{
    SM_PROLOGUE_RC(ss_m::destroy_fixed_rec, in_xct, 0);
    RES_SMSCRIPT(<<"destroy_fixed_rec " << fid << " " << idx);
    W_DO(_destroy_fixed_rec(fid, idx));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::update_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::update_fixed_rec(const rid_t& rid, smsize_t start, const vec_t& data)
{
    SM_PROLOGUE_RC(ss_m::update_fixed_rec, in_xct, 0);
    RES_SMSCRIPT(<<"update_fixed_rec " <<rid <<" "<<start
	<<" [vec " <<data << "] "
    );
    W_DO(_update_fixed_rec(rid, start, data));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::update_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::update_fixed_rec(const stid_t& fid, const uint4 idx, smsize_t start, const vec_t& data)
{
    SM_PROLOGUE_RC(ss_m::update_fixed_rec, in_xct, 0);
    RES_SMSCRIPT(<<"update_fixed_rec " <<fid << " " << idx 
        <<" "<<start <<" [vec " <<data << "] "
    );
    W_DO(_update_fixed_rec(fid, idx, start, data));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::read_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::read_fixed_rec(const stid_t& fid, const uint4 idx, smsize_t start, uint4& len, 
                     void* buf)
{
    SM_PROLOGUE_RC(ss_m::read_fixed_rec, in_xct, 0);
    RES_SMSCRIPT(<<"read_fixed_rec " <<fid << " " << idx <<" "<<start);
    W_DO(_read_fixed_rec(fid, idx, start, len, buf));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::read_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::read_fixed_rec(const rid_t& rid, smsize_t start, uint4& len, 
                     void* buf)
{
    SM_PROLOGUE_RC(ss_m::read_fixed_rec, in_xct, 0);
    RES_SMSCRIPT(<<"read_fixed_rec " <<rid <<" "<<start);
    W_DO(_read_fixed_rec(rid, start, len, buf));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::read_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::read_fixed_rec(const stid_t& fid, const uint4 idx, uint4& len, void* buf)
{
    SM_PROLOGUE_RC(ss_m::read_fixed_rec, in_xct, 0);
    RES_SMSCRIPT(<<"read_fixed_rec " <<fid<< " " <<idx) ; 
    W_DO(_read_fixed_rec(fid, idx, 0, len, buf));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::bulkld_fixed_file()                                   *
 *--------------------------------------------------------------*/
rc_t
ss_m::bulkld_fixed_file( const stid_t& stid, const stid_t& source,
                         sm_du_stats_t& stats)
{
    SM_PROLOGUE_RC(ss_m::bulkld_fixed_file, in_xct, 0);
    SMSCRIPT(<<"bulkld_fixed_file" <<stid<<" "<<source<<" "<<stats);
    W_DO( _bulkld_fixed_file(stid, source, stats) );
    return RCOK;
}

rc_t
ss_m::_bulkld_fixed_file( const stid_t& stid, const stid_t& source,
                          sm_du_stats_t& stats)
{
    FUNC(ss_m::_bulkld_fixed_file);

    sdesc_t* sd = NULL ;

    stpgid_t stpgid(stid);
    W_DO( dir->access(stpgid, sd, EX ) );

    if (sd->sinfo().stype != t_file)   return RC(eBADSTORETYPE);
    W_DO( fl->bulk_load(stid, source, *sd) ) ;  

    sdesc_t temp = *sd ; 
    W_DO( dir->remove(stpgid) ) ;
    W_DO( dir->insert(stpgid,temp.sinfo()) ) ; 

    // TO DO insert the directory entry back
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_create_fixed_file()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_create_fixed_file(vid_t vid, stid_t& fid,
		   store_property_t property, uint4 length)
{
    FUNC(ss_m::_create_fixed_file);
    DBG( << "Attempting to create a file on volume " << vid.vol );

    uint4_t st_flag = _make_store_flag(property);

    DBGTHRD(<<"about to create a store for directory of fixed length");
    W_DO( io->create_store(vid, 100/*unused*/, st_flag, fid) );

    DBGTHRD(<<"created store " << fid);

    // Note: theoretically, some other thread could destroy
    // 	     the above stores before the following lock request
    // 	     is granted.  The only forseable way for this to
    //	     happen would be due to a bug in a vas causing
    //       it to destroy the wrong store.  We make no attempt
    //       to prevent this.
    W_DO(lm->lock(fid, EX, t_long, WAIT_SPECIFIED_BY_XCT));

    DBGTHRD(<<"locked " << fid);

    lpid_t first;
    serial_t  logical_id ; 
    W_DO( fl->create(fid, first, length) );
    DBGTHRD(<<"locked &created " << fid);

    sinfo_s sinfo(fid.store, t_file, 100/*unused*/, 100/*unused*/, t_bad_ndx_t,
		      property, first.page, logical_id, 0, 0);
    sinfo.set_record_size( length ) ; 
    sinfo.set_tree_level( 0 ) ; 
    sinfo.set_free_bitmap( 0 ) ; 

    stpgid_t stpgid(fid);

    W_DO( dir->insert(stpgid, sinfo) );

    DBGTHRD(<<"inserted " << fid.store);

    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_destroy_fixed_file()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_destroy_fixed_file(const stid_t& fid)
{
    FUNC(ss_m::_destroy_fixed_file);

    sdesc_t* sd;

    DBGTHRD(<<"want to destroy store " << fid);

    stpgid_t stpgid(fid);
    W_DO( dir->access(stpgid, sd, EX) );

    // TO DO ... Should be changed to t_fixed store type
    if (sd->sinfo().stype != t_file) return RC(eBADSTORETYPE);

    uint4_t store_flags;
    W_DO( io->get_store_flags(fid, store_flags) );
   
    DBGTHRD(<<"destroying store " << fid);
    W_DO( io->destroy_store(fid) );

    if (store_flags & page_p::st_tmp)  { // for temporary & no-log files, 
	// we can discard the buffers to save
	// some disk I/O.
	W_IGNORE( bf->discard_store(fid) );
        W_IGNORE( bf->discard_store(sd->large_stid()) );
    }
    
    DBGTHRD(<<"about to remove directory entry " << fid);
    W_DO( dir->remove(fid) );
    DBGTHRD(<<"removed directory entry " << fid);
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_create_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_create_fixed_rec( const stid_t& fid, const vec_t& data, rid_t& new_rid) 
{
    return RCOK ; 
}

/*--------------------------------------------------------------*
 *  ss_m::_create_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t     
ss_m::_create_fixed_rec( const stid_t& fid, const vec_t& data, uint4& idx)
{
    return RCOK ; 
}

/*--------------------------------------------------------------*
 *  ss_m::_update_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_update_fixed_rec(const rid_t& rid, smsize_t start, const vec_t& data)
{
    // TO DO W_DO(lm->lock(rid, EX, t_long, WAIT_SPECIFIED_BY_XCT));
    W_DO(fl->update_rec(rid, start, data));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_update_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_update_fixed_rec(const stid_t& fid, const uint4 idx, smsize_t start, 
                        const vec_t& data)
{
    // Get the directory information 
    sdesc_t* sd;
    W_DO( dir->access(fid, sd, EX) );

    W_DO( fl->update_rec( sd->root(), idx, sd->sinfo().record_size, 
                         sd->sinfo().tree_level, start, data) );
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_destroy_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_destroy_fixed_rec(const rid_t& rid)
{
    // TO DO W_DO(lm->lock(rid, EX, t_long, WAIT_SPECIFIED_BY_XCT));
    W_DO(fl->destroy_rec(rid));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_destroy_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_destroy_fixed_rec(const stid_t& fid, const uint4 idx)
{
    // Get the directory information 
    sdesc_t* sd;
    W_DO( dir->access(fid, sd, SH) );

    W_DO( fl->destroy_rec(sd->root(), idx, sd->sinfo().record_size,
                         sd->sinfo().tree_level ));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_read_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_read_fixed_rec(const rid_t& rid, smsize_t start, uint4& len, 
                      void* buf)
{
    // TO DO... mapping of the idx within the lock manager
    // W_DO(lm->lock(rid, SH, t_long, WAIT_SPECIFIED_BY_XCT));

    W_DO(fl->read_rec(rid, start, len, buf));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_read_fixed_rec()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_read_fixed_rec(const stid_t& fid, const uint4 idx, smsize_t start, uint4& len, 
                      void* buf)
{
    // Get the directory information 
    sdesc_t* sd;
    W_DO( dir->access(fid, sd, SH) );

    W_DO( fl->read_rec(sd->root(), idx, sd->sinfo().record_size,
                      sd->sinfo().tree_level, start, len, buf) );

    return RCOK;
}
