/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: smfile.c,v 1.29 1996/10/28 21:45:36 nhall Exp $
 */
#define SM_SOURCE
#define SMPLAIN_C

#include "w.h"
#include "option.h"
#include "sm_int.h"
#include "device.h"
#include "vol.h"
#include "app_support.h"
#include "remote_client.h"

#define DBGTHRD(arg) DBG(<<" th."<<me()->id<<" tid."<<xct()->tid()<<" " arg)

/*==============================================================*
 *  Physical ID version of all the storage operations           *
 *==============================================================*/

/*--------------------------------------------------------------*
 *  ss_m::create_plain_file()				        *
 *--------------------------------------------------------------*/
rc_t
ss_m::create_plain_file(
    vid_t 			vid, 
    stid_t& 			fid, 
    store_property_t 		property,
    uint4  		        length)
{
    SM_PROLOGUE_RC(ss_m::create_plain_file, in_xct, 0);
    RES_SMSCRIPT(<<"create_plain_file " <<vid << " " << property << " " << length );
    W_DO(_create_plain_file(vid, fid, property, length));
    DBGTHRD(<<"create_plain_file returns " << fid);
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::destroy_plain_file()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::destroy_plain_file(const stid_t& fid)
{
    SM_PROLOGUE_RC(ss_m::destroy_plain_file, in_xct, 0);
    RES_SMSCRIPT(<<"destroy_plain_file " <<fid);
    W_DO(_destroy_plain_file(fid));
    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_create_plain_file()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_create_plain_file(vid_t vid, stid_t& fid,
		   store_property_t property, uint4 length)
{
    FUNC(ss_m::_create_plain_file);
    DBG( << "Attempting to create a file on volume " << vid.vol );

    uint4_t st_flag = _make_store_flag(property);

    DBGTHRD(<<"about to create a store for directory of fixed length");
    W_DO( io->create_store(vid, 100/*unused*/, st_flag, fid) );

    DBGTHRD(<<"created store " << fid);

    // Note: theoretically, some other thread could destroy
    // 	     the above stores before the following lock request
    // 	     is granted.  The only forseable way for this to
    //	     happen would be due to a bug in a vas causing
    //       it to destroy the wrong store.  We make no attempt
    //       to prevent this.
    W_DO(lm->lock(fid, EX, t_long, WAIT_SPECIFIED_BY_XCT));

    DBGTHRD(<<"locked " << fid);

    lpid_t first;
    serial_t  logical_id ; 
    W_DO( fl->create(fid, first, length) );
    DBGTHRD(<<"locked &created " << fid);

    sinfo_s sinfo(fid.store, t_file, 100/*unused*/, 100/*unused*/, t_bad_ndx_t,
		      property, first.page, logical_id, 0, 0);
    sinfo.set_record_size( length ) ; 
    sinfo.set_last_page( first.page ) ; 

    stpgid_t stpgid(fid);

    W_DO( dir->insert(stpgid, sinfo) );

    DBGTHRD(<<"inserted " << fid.store);

    return RCOK;
}

/*--------------------------------------------------------------*
 *  ss_m::_destroy_plain_file()					*
 *--------------------------------------------------------------*/
rc_t
ss_m::_destroy_plain_file(const stid_t& fid)
{
    FUNC(ss_m::_destroy_plain_file);

    sdesc_t* sd;

    DBGTHRD(<<"want to destroy store " << fid);

    stpgid_t stpgid(fid);
    W_DO( dir->access(stpgid, sd, EX) );

    // TO DO ... Should be changed to t_plain store type
    if (sd->sinfo().stype != t_file) return RC(eBADSTORETYPE);

    uint4_t store_flags;
    W_DO( io->get_store_flags(fid, store_flags) );
   
    DBGTHRD(<<"destroying store " << fid);
    W_DO( io->destroy_store(fid) );

    if (store_flags & page_p::st_tmp)  { // for temporary & no-log files, 
	// we can discard the buffers to save
	// some disk I/O.
	W_IGNORE( bf->discard_store(fid) );
        W_IGNORE( bf->discard_store(sd->large_stid()) );
    }
    
    DBGTHRD(<<"about to remove directory entry " << fid);
    W_DO( dir->remove(fid) );
    DBGTHRD(<<"removed directory entry " << fid);
    return RCOK;
}

