/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: smthread.c,v 1.38 1996/11/01 23:44:35 bolo Exp $
 */
#define SM_SOURCE
#define SMTHREAD_C
#ifdef __GNUG__
#   pragma implementation
#endif

#include <sm_int.h>
//#include <e_error.i>

const eERRMIN = smlevel_0::eERRMIN;
const eERRMAX = smlevel_0::eERRMAX;

int smthread_init_t::count = 0;


/*
 *  smthread_init_t::smthread_init_t()
 */
smthread_init_t::smthread_init_t()
{
}



/*
 *	smthread_init_t::~smthread_init_t()
 */
smthread_init_t::~smthread_init_t()
{
}

/*********************************************************************
 *
 *  Constructor and destructor for smthread_t::tcb_t
 *
 *********************************************************************/

void
smthread_t::no_xct()
{
    if(tcb()._sdesc_cache) {
	xct_t::delete_sdesc_cache_t(tcb()._sdesc_cache);
	tcb()._sdesc_cache =0;
    }
    if(tcb()._lock_hierarchy) {
	xct_t::delete_lock_hierarchy(tcb()._lock_hierarchy);
	tcb()._lock_hierarchy =0;
    }
}
void
smthread_t::new_xct()
{
    tcb()._sdesc_cache = xct_t::new_sdesc_cache_t();
    tcb()._lock_hierarchy = xct_t::new_lock_hierarchy();
}



/*********************************************************************
 *
 *  smthread_t::smthread_t
 *
 *  Create an smthread_t.
 *
 *********************************************************************/
smthread_t::smthread_t(
    st_proc_t* f,
    void* arg,
    priority_t priority,
    bool block_immediate,
    bool auto_delete,
    const char* name,
    long lockto)
: sthread_t(priority, block_immediate, auto_delete, name),
  _proc(f),
  _arg(arg)
#ifndef OLD_SM_BLOCK
  ,_block("smblock"),
  _awaken("smawaken")
#endif
{
	lock_timeout(lockto);
}


smthread_t::smthread_t(
    priority_t priority,
    bool block_immediate,
    bool auto_delete,
    const char* name,
    long lockto)
: sthread_t(priority, block_immediate, auto_delete, name),
  _proc(0),
  _arg(0)
#ifndef OLD_SM_BLOCK
  ,_block("smblock"),
  _awaken("smawaken")
#endif
{
	lock_timeout(lockto);
}




/*********************************************************************
 *
 *  smthread_t::~smthread_t()
 *
 *  Destroy smthread. Thread is already defunct the object is
 *  destroyed.
 *
 *********************************************************************/
smthread_t::~smthread_t()
{
    // w_assert3(tcb().pin_count == 0);
    no_xct();
}

#ifndef OLD_SM_BLOCK
w_rc_t	smthread_t::block(smutex_t &lock,
			  int4_t timeout,
			  const char * const)
{
	bool	timed_out = false;

	W_COERCE(lock.acquire());
	_unblocked = false;
	_waiting = true;

	/* XXX adjust timeout for "false" signals */
	while (!_unblocked && !timed_out) {
		w_rc_t	e = _awaken.wait(lock, timeout);
		if (e && e.err_num() == stTIMEOUT)
			timed_out = true;
		else if (e)
			W_COERCE(e);
	}

	_waiting = false;
	lock.release();

	/* XXX possible race condition on sm_rc, except that
	   it DOES belong to me, this thread?? */
	/* XXX if so, the thread package _rc code has the same problem */
	return timed_out ? RC(stTIMEOUT) : _sm_rc;
}

w_rc_t	smthread_t::unblock(smutex_t &lock, const w_rc_t &rc)
{
	W_COERCE(lock.acquire());
	if (!_waiting) {
		cerr << "Warning: smthread_t::unblock():"
			<< " thread not waiting, unblock lost" << endl;
	}

	_unblocked = true;

	/* Save rc (will be returned by block());
	   this code is copied from that in the
	   sthread block/unblock sequence. */
	if (_sm_rc)  {;}
	if (&rc) 
		_sm_rc = rc;
	else
		_sm_rc = RCOK;

	_awaken.signal();
	lock.release();
	return RCOK;
}

/* thread-compatability block() and unblock.  Use the per-smthread _block
   as the synchronization primitive. */
w_rc_t	smthread_t::block(int4_t timeout,
			  sthread_list_t *list,
			  const char * const caller,
			  const void * id)
{
	w_assert1(list == 0);
	return block(_block, timeout, caller);
}

w_rc_t	smthread_t::unblock(const w_rc_t &rc)
{
	return unblock(_block, rc);
}
#endif


/*********************************************************************
 *
 *  smthread_t::run()
 *
 *  Befault body of smthread. Could be overriden by subclass.
 *
 *********************************************************************/
#ifdef NOTDEF
// Now this function is pure virtual in an attempt to avoid
// a possible gcc bug

void smthread_t::run()
{
    w_assert1(_proc);
    _proc(_arg);
}
#endif /* NOTDEF */


void 
smthread_t::attach_xct(xct_t* x)
{
    w_assert3(tcb().xct == 0); 
    tcb().xct = x;
    int n = x->attach_thread();
    w_assert1(n >= 1);
}


void 
smthread_t::detach_xct(xct_t* x)
{
    w_assert3(tcb().xct == x); 
    int n=x->detach_thread();
    w_assert1(n >= 0);
    tcb().xct = 0;
}

void		
smthread_t::_dump(ostream &o)
{
	sthread_t *t = (sthread_t *)this;
	t->sthread_t::_dump(o);

	o << "smthread_t: " << (char *)(is_in_sm()?"in sm ":"");
	if(tcb().xct) {
	  o << "xct " << tcb().xct->tid() << endl;
	}
// no output operator yet
//	if(sdesc_cache()) {
//	  o << *sdesc_cache() ;
//	}
	o << endl;
}
